/* sane - Scanner Access Now Easy.
   Copyright (C) 1997 David Mosberger-Tang/Peter Kirchgessner
   This file is just for testing HP-Scanners for supported commands.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.

   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.

   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.

   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice.

   This file implements an inquiry tool to test the supported SCL-commands
   for HP-scanner.

   *NO WARRANTY*   */

/* Changes:
   12-Dec-98, print SCSI-inquire string, compatibilities
   01-Nov-98, added some more SCL-commands
              Tried to test supported ranges of certain commands by issueing
              the command with that value. But the scanner does not set an
              error condition for unsupported values.
   26-Apr-98, added some more SCL-commands
   09-Sep-00, added PTAL-support (David Paschal)
   08-Oct-00, add direct device support (USB)
   06-Oct-02, add Button State
   14-Jan-03, more detailed debug logs
*/

#include "sane/config.h"

#include <limits.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "sane/sane.h"
#include "sane/saneopts.h"
#include "sane/sanei_scsi.h"

#define BACKEND_NAME	hpinq
#include "sane/sanei_backend.h"

#ifndef PATH_MAX
# define PATH_MAX	1024
#endif

#include "sane/sanei_config.h"

#include "hptools.h"
#define TEST_ALL_COMMANDS

#define INQ_LEN	255
static const u_int8_t inquiry[] =
  {
    0x12, 0x00, 0x00, 0x00, INQ_LEN, 0x00
  };

static const u_int8_t test_unit_ready[] =
  {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00
  };

static char *str_to_oct (const char *input, int len)
{
 static unsigned char oct[2048];
 unsigned char *dst = oct, *lastchar = &(oct[sizeof (oct)-1]);
 int k, ch;

 for (k = 0; k < len; k++)
 {
   ch = ((const unsigned char *)input)[k];
   if ((ch < ' ') || (ch >= 127))
   {
     if (lastchar - dst >= 4)
     {
       sprintf ((char *)dst, "\\%03o", ch);
       dst+= 4;
     }
   }
   else
   {
     if (dst < lastchar) *(dst++) = (unsigned char)ch;
   }
 }
 *dst = '\0';
 return (char *)oct;
}


static SANE_Status
read_data (int fd, u_char * dest, size_t *len)
{
  static u_char read_6[6];

  memset (read_6, 0, sizeof (read_6));
  read_6[0] = 0x08;
  read_6[2] = (*len >> 16) & 0xff;
  read_6[3] = (*len >> 8) & 0xff;
  read_6[4] = (*len) & 0xff;
  return hptools_device_cmd (fd, read_6, sizeof (read_6), dest, len);
}

static SANE_Status
write_data (int fd, void *data, size_t len)
{
  char write_6[4096];

  memset (write_6, 0, 6);
  write_6[0] = 0xa;
  write_6[2] = len >> 16;
  write_6[3] = len >> 8;
  write_6[4] = len;
  memcpy (write_6 + 6, data, len);

  return hptools_device_cmd (fd, (unsigned char *)write_6, 6 + len, 0, 0);
}

static SANE_Status
write_string (int fd, const char *format, ...)
{
  va_list args;
  char buf[1024];

  va_start (args, format);
  vsprintf (buf, format, args);
  return write_data (fd, buf, strlen (buf));
}

/* return values: 0 = ok, 1 = got null response, 2 = got malformed response
 *                3 = read-error
 */
static int
inquire (int fd, int request_code, char request_letter,
	 char reply_letter, int *valp)
{
  u_char buf[64], letter, lastchar;
  SANE_Status status;
  int code;
  size_t len;

  *valp = 0;

  sprintf ((char *) buf, "\033*s%d%c", request_code, request_letter);
  write_string (fd, (char *) buf);
  len = sizeof (buf);
  memset (buf, 0, sizeof (buf));
  status = read_data (fd, buf, &len);
  if (status != SANE_STATUS_GOOD)
    {
      DBG(1, "inquire: read_data failed (%s)\n", sane_strstatus (status));
      return 3;
    }

  DBG(3, "inquire: read <%s>\n", str_to_oct ((char *)buf, len));

  if (sscanf ((char *) buf, "\033*s%d%c%d%c", &code, &letter, valp, &lastchar)
      == 4)
  {
    if ((code == request_code) && (letter == reply_letter) && (lastchar == 'V'))
    {
      return 0;
    }
    else
    {
      DBG(1, "inquire: returned string is malformed (%s)\n", buf);
      return 2;
    }
  }
  if (sscanf ((char *) buf, "\033*s%d%c%c", &code, &letter, &lastchar)
      == 3)
  {
    if ((code == request_code) && (letter== reply_letter) && (lastchar == 'N'))
    {
      DBG(5, "inquire: returned null response (%s)\n", buf);
      return (1);
    }
  }
 DBG(1, "inquire: returned string is malformed (%s)\n", buf);
 return 2;
}


/* return values: 0 = ok, 1 = got null response, 2 = got malformed response
 *                3 = read-error
 */
static int
inquire_cmd (int fd, int inq_id, int *supported,
             int *minval, int *maxval, int *cuval)
{int retval;

 *supported = 0;

 retval = inquire (fd, inq_id, 'R', 'p', cuval);
 if (retval != 0) return (retval);

 retval = inquire (fd, inq_id, 'L', 'k', minval);
 if (retval != 0) return (retval);

 retval = inquire (fd, inq_id, 'H', 'g', maxval);
 if (retval != 0) return (retval);

 *supported = 1;

 return 0;
}


/* return values: 0 = ok, 1 = got null response, 2 = got malformed response
 *                3 = read-error
 */
static int
inquire_devparam (int fd, int inq_id, char *response)

{u_char buf[64], null_response[32];
 SANE_Status status;
 size_t len, null_len;
 int i_id, i_val;
 char i_char;

 sprintf ((char *) buf, "\033*s%dE", inq_id);
 sprintf ((char *) null_response, "\033*s%ddN", inq_id);
 null_len = strlen ((char *)null_response);

 DBG (3, "inquire_devparam: write <ESC>%s\n", buf+1);

 write_string (fd, (char *) buf);
 len = sizeof (buf);
 memset (buf, 0, len);
 status = read_data (fd, buf, &len);
 if (status != SANE_STATUS_GOOD)
 {
   DBG(1, "inquire_devparam: read_data failed (%s)\n",
          sane_strstatus (status));
   return 3;
 }

 DBG (3, "inquire_devparam: read <%s>\n", str_to_oct ((char *)buf, len));

 if (strncmp (buf, null_response, null_len) == 0)
   return 1;

 if (sscanf ((char *)buf, "\033*s%dd%d%c", &i_id, &i_val, &i_char) != 3)
 {
   DBG(1, "inquire_devparam: could not decode response (%s)\n", buf);
   return 2;
 }
 if (i_char == 'V')
   sprintf (response, "%d", i_val);
 else if (i_char == 'W')
 {
   strncpy (response, index (buf, 'W')+1, i_val);
   response[i_val] = '\0';
 }
 else
 {
   strncpy (response, buf, sizeof (buf));
   response[sizeof (buf) - 1] = '\0';
 }

 return 0;
}


#ifdef TEST_ALL_VALUES
/* Set a command with a value */
static SANE_Status
set_value (int fd, int inq_id, int value)

{int c1, c2;
 char buf[64];

 c1 = (inq_id - 10306) / 32 + 'a';
 c2 = (inq_id - 10306) % 32 + 'A';
 sprintf (buf, "\033*%c%d%c", c1, value, c2);  /* Compose command */
 return (write_string (fd, buf));
}


/* Test if a certain value of a command is supported */
static int
test_value (int fd, int inq_id, int value)

{int c1, c2;
 char buf[64];

 write_string (fd, "\033*oE");      /* Clear errors */
 c1 = (inq_id - 10306) / 32 + 'a';
 c2 = (inq_id - 10306) % 32 + 'A';
 sprintf (buf, "\033*%c%d%c", c1, value, c2);  /* Compose command */

 DBG(2, "test_value: write <ESC>%s\n", buf+1);

 write_string (fd, buf);
                                 /* Inquire error stack */
 if ((inquire_devparam (fd, 257, buf) == 0) && (buf[0] == '0'))
   return (0);
 else
   return (1);
}
#endif


static char *
hp_model (int fd, int list_all)

{char response[256];
 static char retval[256];
 int k;
 static struct {
   int id;
   char *name;
 }
 model[] =
 {
   { 32, "HP ScanJet 6300C/6350C" },
   { 31, "HP ScanJet 5200C" },
   { 21, "HP ScanJet 6200C/6250C" },
   { 19, "HP OfficeJet 1170C or later" },
   { 18, "HP OfficeJet 1150C" },
   { 17, "HP PhotoSmart Photo Scanner" },
   { 16, "HP ScanJet 5p/4100C/5100C" },
   { 15, "HP ScanJet 4p" },
   { 14, "HP ScanJet 3p" },
   { 12, "HP ScanJet 3c/4c/6100C" },
   { 11, "HP ScanJet IIcx" },
   { 9,  "HP ScanJet IIp" },
   { 10, "HP ScanJet IIc" },
   { 3,  "HP ScanJet/Plus" }
 };

 for (k = 0; k < NELEMS (model); k++)
 {
  if (inquire_devparam (fd, model[k].id, response) == 0)
  {
    if (!list_all)
    {
      sprintf (retval, "%s (%s)", model[k].name, response);
      return retval;
    }
    printf ("%s\n", model[k].name);
  }
 }
 return "UNKNOWN";
}


static SANE_Status
hp_inquire (const char *devname, HpConnect connect)
{
  char result[INQ_LEN], response[256], *dev_param;
  unsigned char *uc;
  static char *errstr[] = { "ok", "n.a. (null response)",
     "n.a. (malformed response)", "n.a. (error in inquiry)" };
  int k, i, fd, ev;
  int supported, minval, maxval, cuval;
  SANE_Status status;
  size_t size;
  static struct {
    char *command;
    char *code;
    int inq_id;
    int test_range;
  }
#ifdef TEST_ALL_COMMANDS
  *inq_cmd_rec,
#endif
  inq_cmd[] =
  {
#ifndef TEST_ALL_COMMANDS
    { NULL,                     "Esc*a#A", 10306, 0 },
#endif
    { "Automatic Bkgrd",        "Esc*a#B", 10307, 0 },
    { "Compression",            "Esc*a#C", 10308, 1 },
    { "Download Type",          "Esc*a#D", 10309, 1 },
    { "X Scale",                "Esc*a#E", 10310, 0 },
    { "Y Scale",                "Esc*a#F", 10311, 0 },
    { "Select Data Width",      "Esc*a#G", 10312, 0 }, /*Depends on data type*/
#ifndef TEST_ALL_COMMANDS
    { NULL,                     "Esc*a#H", 10313, 0 },
#endif
    { "Inverse Image",          "Esc*a#I", 10314, 0 },
    { "B/W Dither Pattern",     "Esc*a#J", 10315, 1 },
    { "Contrast Level",         "Esc*a#K", 10316, 0 },
    { "Intensity Level",        "Esc*a#L", 10317, 0 },
    { "Mirror Image",           "Esc*a#M", 10318, 0 },
#ifndef TEST_ALL_COMMANDS
    { NULL,                     "Esc*a#N", 10319, 0 },
#endif
    { "Reserved1",              "Esc*a#O", 10320, 0 }, /* scl.c */
    { "Window X Ext (Dec)",     "Esc*a#P", 10321, 0 },
    { "Window Y Ext (Dec)",     "Esc*a#Q", 10322, 0 },
    { "X Res",                  "Esc*a#R", 10323, 0 },
    { "Y Res",                  "Esc*a#S", 10324, 0 },
    { "Output Data Type",       "Esc*a#T", 10325, 1 },
#ifndef TEST_ALL_COMMANDS
    { NULL,                     "Esc*a#U", 10326, 0 },
#endif
    { "Scan speed",             "Esc*a#V", 10327, 0 }, /* scl.c */
    { "Download Binary",        "Esc*a#W[]", 10328, 0 },
    { "Window X Pos (Dec)",     "Esc*a#X", 10329, 0 },
    { "Window Y Pos (Dec)",     "Esc*a#Y", 10330, 0 },
    { "Reconnect level",        "Esc*a#Z", 10331, 0 }, /* scl.c */
    { "Select Number of Views", "Esc*f#A", 10466, 0 },
    { "Select View",            "Esc*f#B", 10467, 0 },
    { "Preload ADF",            "Esc*f#C", 10468, 0 },
    { "Media type",             "Esc*f#D", 10469, 0 }, /* scl.c */
    { "Coordinate system",      "Esc*f#E", 10470, 0 }, /* scl.c */
    { "Pos Secondary Elem",     "Esc*f#F", 10471, 0 },
#ifndef TEST_ALL_COMMANDS
    { NULL,                     "Esc*f#G", 10472, 0 },
    { NULL,                     "Esc*f#H", 10473, 0 },
    { NULL,                     "Esc*f#I", 10474, 0 },
    { NULL,                     "Esc*f#J", 10475, 0 },
    { NULL,                     "Esc*f#K", 10476, 0 },
#endif
    { "Light Source Test",      "Esc*f#L", 10477, 0 },
#ifndef TEST_ALL_COMMANDS
    { NULL,                     "Esc*f#M", 10478, 0 },
    { NULL,                     "Esc*f#N", 10479, 0 },
    { NULL,                     "Esc*f#O", 10480, 0 },
#endif
    { "Window X Ext (Pix)",     "Esc*f#P", 10481, 0 },
    { "Window Y Ext (Pix)",     "Esc*f#Q", 10482, 0 },
#ifndef TEST_ALL_COMMANDS
    { NULL,                     "Esc*f#R", 10483, 0 },
#endif
    { "Scan Window (w/ID)",     "Esc*f0S", 10484, 0 },
    { "Scanner Test",           "Esc*f#T", 10485, 0 }, /* scl.c */
#ifndef TEST_ALL_COMMANDS
    { NULL,                     "Esc*f#U", 10486, 0 },
    { NULL,                     "Esc*f#V", 10487, 0 },
    { NULL,                     "Esc*f#W", 10488, 0 },
#endif
    { "Window X Pos (Pix)",     "Esc*f#X", 10489, 0 },
    { "Window Y Pos (Pix)",     "Esc*f#Y", 10490, 0 },
#ifndef TEST_ALL_COMMANDS
    { NULL,                     "Esc*f#Z", 10491, 0 },
    { NULL,                     "Esc*s#A", 10882, 0 },
    { NULL,                     "Esc*s#B", 10883, 0 },
    { NULL,                     "Esc*s#C", 10884, 0 },
    { NULL,                     "Esc*s#D", 10885, 0 },
    { NULL,                     "Esc*s#E", 10886, 0 },
    { NULL,                     "Esc*s#F", 10887, 0 },
    { NULL,                     "Esc*s#G", 10888, 0 },
    { NULL,                     "Esc*s#H", 10889, 0 },
    { NULL,                     "Esc*s#I", 10890, 0 },
    { NULL,                     "Esc*s#J", 10891, 0 },
    { NULL,                     "Esc*s#K", 10892, 0 },
    { NULL,                     "Esc*s#L", 10893, 0 },
    { NULL,                     "Esc*s#M", 10894, 0 },
    { NULL,                     "Esc*s#N", 10895, 0 },
    { NULL,                     "Esc*s#O", 10896, 0 },
    { NULL,                     "Esc*s#P", 10897, 0 },
    { NULL,                     "Esc*s#Q", 10898, 0 },
    { NULL,                     "Esc*s#R", 10899, 0 },
    { NULL,                     "Esc*s#S", 10900, 0 },
    { NULL,                     "Esc*s#T", 10901, 0 },
#endif
    { "Upload Binary",          "Esc*s#U", 10902, 1 },
#ifndef TEST_ALL_COMMANDS
    { NULL,                     "Esc*s#V", 10903, 0 },
    { NULL,                     "Esc*s#W", 10904, 0 },
    { NULL,                     "Esc*s#X", 10905, 0 },
    { NULL,                     "Esc*s#Y", 10906, 0 },
    { NULL,                     "Esc*s#Z", 10907, 0 },
#endif
    { "Select Cal Area (Y-pos)","Esc*u#A", 10946, 0 },
    { "Byte order",             "Esc*u#B", 10947, 0 },
    { "Select Cal Strip Parm.", "Esc*u#C", 10948, 0 },
    { "XPA Scan Window (w/ID)", "Esc*u0D", 10949, 0 },
    { "Speed Mode",             "Esc*u#E", 10950, 1 },
    { "Set Filter",             "Esc*u#F", 10951, 1 },
    { "Calibration mode",       "Esc*u#G", 10952, 0 }, /* scl.c */
    { "XPA Disable/Enable",     "Esc*u#H", 10953, 0 },
#ifndef TEST_ALL_COMMANDS
    { NULL,                     "Esc*u#I", 10954, 0 },
#endif
    { "Color Dither Pattern",   "Esc*u#J", 10955, 1 },
    { "Select Tone Map",        "Esc*u#K", 10956, 1 },
    { "Light Power",            "Esc*u#L", 10957, 0 }, /* scl.c */
#ifndef TEST_ALL_COMMANDS
    { NULL,                     "Esc*u#M", 10958, 0 },
    { NULL,                     "Esc*u#N", 10959, 0 },
    { NULL,                     "Esc*u#O", 10960, 0 },
    { NULL,                     "Esc*u#P", 10961, 0 },
    { NULL,                     "Esc*u#Q", 10962, 0 },
#endif
    { "Recalibrate (w/ID)",     "Esc*u0R", 10963, 0 },
    { "ADF Scan Window (w/ID)", "Esc*u0S", 10964, 0 },
    { "Select Matrix Coef",     "Esc*u#T", 10965, 1 },
    { "Unload Document",        "Esc*u0U", 10966, 0 },
#ifndef TEST_ALL_COMMANDS
    { NULL,                     "Esc*u#V", 10967, 0 },
    { NULL,                     "Esc*u#W", 10968, 0 },
#endif
    { "Change Document",        "Esc*u0X", 10969, 0 },
    { "ADF Bkgrd Feed",         "Esc*u0Y", 10970, 0 },
#ifndef TEST_ALL_COMMANDS
    { NULL,                     "Esc*u#Z", 10971, 0 }
#endif
  };
  static struct {
    int inq_id;
    char *dev_param;
  }
  inq_devparam[] =
  {
    {    3, "HP Model #1" },
    {   10, "HP Model #2" },
    {    9, "HP Model #3" },
    {   11, "HP Model #4" },
    {   12, "HP Model #5" },
    {   14, "HP Model #6" },
    {   15, "HP Model #8" },
    {   16, "HP Model #9" },
    {   17, "HP Model #10" },
    {   18, "HP Model #11" },
    {   19, "HP Model #12" },
    {   21, "HP Model #14" },
    {   31, "HP Model #16" },
    {   32, "HP Model #17" },
    {    4, "Firmware Date" },
    {    5, "Scan Status" },     /* scl.c */
    {    6, "ADF Status" },      /* scl.c */
    {   21, "Reserved" },        /* scl.c */
    {   24, "ADF Capability" },
    {   25, "ADF Bin" },
    {   26, "Open Since last Chg" },
    {   27, "Ready to Unload" },
    {   28, "XPA Capability" },
    {   29, "XPA Status" },      /* scl.c */
    {   30, "XPA Ready" },
    {  256, "Max Error Stack" },
    {  257, "Current Error Stack" },
    {  259, "Current Error Number" },
    {  261, "Oldest Error" },
    { 1024, "Pixels/Scan Line" },
    { 1025, "Bytes/Scan Line" },
    { 1026, "Number Scan Lines" },
    { 1027, "ADF Ready" },
    { 1028, "Device Pixels/Inch" },
    { 1029, "Native Optical Resolution" },
    { 1031, "Auto Filtering" },
    { 1034, "Should Be" },            /* scl.c */
    { 1037, "Bytes in last scan" },   /* scl.c */
    { 1040, "Inquire Auto Speed Mode" },
    { 1041, "Which ADF" },            /* scl.c */
    { 1044, "Button State" },         /* undocumented */
    { 1045, "Inquire Preloaded" },
    { 1046, "Auto coefficient" },     /* scl.c */
    { 1047, "Scan direction" }        /* scl.c */
  };
#ifdef TEST_ALL_COMMANDS
	unsigned char a='*',b,c;
	char *command;
	int inq_id,n=NELEMS(inq_cmd);
#endif


  DBG_INIT ();

  DBG(3, "hp_inquire: opening %s\n", devname);
  status = hptools_device_open (devname, connect, &fd, 0, 0);
  if (status != SANE_STATUS_GOOD)
    {
      DBG(1, "hp_inquire: open failed (%s)\n", sane_strstatus (status));
      return SANE_STATUS_INVAL;
    }

  DBG(3, "hp_inquire: sending INQUIRY\n");
  size = sizeof (result);
  status = hptools_device_cmd (fd, inquiry, sizeof (inquiry),
                               (unsigned char *)result, &size);
  if (status != SANE_STATUS_GOOD || size < 16)
    {
      DBG(1, "hp_inquire: inquiry failed (%s)\n", sane_strstatus (status));
      hptools_device_close (fd);
      return status;
    }

  result[36] = '\0';

  DBG(3, "hp_inquire: sending TEST_UNIT_READY\n");
  status = hptools_device_cmd (fd, test_unit_ready, sizeof (test_unit_ready),
			   0, 0);
  if (status != SANE_STATUS_GOOD)
    {
      DBG(1, "hp_inquire: test unit ready failed (%s)\n",
	  sane_strstatus (status));
      hptools_device_close (fd);
      return status;
    }

  printf ("SCL Command summary for %s\n", hp_model (fd, 0));
  printf ("(Manufacturer Model FirmwareCode = %s)\n", &(result[8]));
  printf ("================================================================\n");
  printf ("\n");
  printf ("Command                 Code        Inq.ID    min    max current\n");
  printf ("----------------------------------------------------------------\n");

#ifdef TEST_ALL_COMMANDS
	for (b='a';b<='z';b++) {
	    for (c='A';c<='Z';c++) {
		inq_id=(a-'!'+1)*1024 + (b-'`'+1)*32 + (c-'@'+1);

		command=NULL;
		inq_cmd_rec=NULL;
		for (i=0,inq_cmd_rec=inq_cmd;i<n;i++,inq_cmd_rec++) {
			if (inq_cmd_rec->inq_id==inq_id) {
				command=inq_cmd_rec->command;
				break;
			}
		}

		ev=inquire_cmd(fd,inq_id,&supported,&minval,&maxval,&cuval);
		if (ev==1) inq_cmd_rec->test_range=0;
		/* Don't list unknown commands which are not supported */
		if ((command==NULL) && (ev==1)) continue;

		printf("%-23s Esc%c%c#%c%s   %6d ",
			(command==NULL)?"---?---":command,
			a,b,c,inq_id==10328?"[]":"  ",
			inq_id);
		if ((ev==0) && supported) {
			printf ("%6d %6d %6d", minval, maxval, cuval);
		} else {
			printf ("   %s", errstr[ev]);
		}
		printf ("\n");
	    }
	}

#else
for (i = 0; i < NELEMS (inq_cmd); i++)
  {
    ev = inquire_cmd (fd, inq_cmd[i].inq_id, &supported, &minval, &maxval,
                      &cuval);
    if (ev == 1) inq_cmd[i].test_range = 0;
    /* Don't list unknown commands which are not supported */
    if ((inq_cmd[i].command == NULL) && (ev == 1)) continue;
    printf ("%-23s %-11s %6d ", (inq_cmd[i].command == NULL) ?
            "---?---" : inq_cmd[i].command, inq_cmd[i].code,
            inq_cmd[i].inq_id);
    if ((ev == 0) && supported)
      printf ("%6d %6d %6d", minval, maxval, cuval);
    else
      printf ("   %s", errstr[ev]);
    printf ("\n");
  }
#endif

#ifdef TEST_ALL_VALUES
  /* Here I tried to test each supported value of a command. */
  /* But the scanner (HP Photosmart) did not set an error */
  /* condition for commands with illegal values. */

  printf ("\n\nSupported range for certain commands\n\
------------------------------------\n");
  for (i = 0; i < NELEMS (inq_cmd); i++)
  {int j;
    if (!inq_cmd[i].test_range) continue;
    ev = inquire_cmd (fd, inq_cmd[i].inq_id, &supported, &minval, &maxval,
                      &cuval);
    if (ev != 0) continue;
    printf ("%-23s ", (inq_cmd[i].command == NULL) ?
            "---?---" : inq_cmd[i].command);
    for (j = minval; j <= maxval+1; j++)
    {
      if (test_value (fd, inq_cmd[i].inq_id, j) == 0)
        printf ("%d ", j);
    }
    set_value (fd, inq_cmd[i].inq_id, cuval); /* Restore value */
    printf ("\n");
  }
#endif

  printf ("\n\nDevice Parameters\n==================\n\n");
  printf ("Inquire Device Parameter            Inq.ID    Response\n");
  printf ("-------------------------------------------------------\n");

  for (i = 0; i < 1200; i++)
  {
    ev = inquire_devparam (fd, i, response);
    dev_param = NULL;
    for (k = 0; k < NELEMS (inq_devparam); k++)
    {
      if (i == inq_devparam[k].inq_id)
      {
        dev_param = inq_devparam[k].dev_param;
        break;
      }
    }
    /* Dont list unknown commands which are not supported */
    if ((dev_param == NULL) && (ev == 1)) continue;
    
    /* Suppress non printable characters */
    if (ev == 0)
    {
      uc = (unsigned char *)response;
      for (k = 0; (k < (int)sizeof (response)) && (uc[k] != '\0'); k++)
        if ((uc[k] < 32) || (uc[k] > 127)) uc[k] = ' ';
    }
    printf ("%-36s%6d    %s\n", (dev_param == NULL) ? "---?---" : dev_param, i,
            (ev == 0) ? response : errstr[ev]);
  }

  printf ("\n\nSupported command sets\n======================\n\n");
  hp_model (fd, 1);

  hptools_device_close (fd);

  return SANE_STATUS_GOOD;
}


int main (int argc, char *argv[])

{HpConnect connect;

 if (argc < 2)
 {
   fprintf (stderr, "Usage: %s <devicename> [<connect>]\n", argv[0]);
   fprintf (stderr, "  <connect> : connect-scsi|connect-ptal|connect-device\n");
   exit (1);
 }

 connect = HP_CONNECT_SCSI;
 if (argc > 2)
 {
   if (strcmp (argv[2], "connect-scsi") == 0)
     connect = HP_CONNECT_SCSI;
   else if (strcmp (argv[2], "connect-ptal") == 0)
     connect = HP_CONNECT_PTAL;
   else if (strcmp (argv[2], "connect-device") == 0)
     connect = HP_CONNECT_DEVICE;
   else
     fprintf (stderr, "Warning: unknown connect %s. Assume connect-scsi\n",
              argv[2]);
 }
 hp_inquire (argv[1], connect);

 exit (0);
}
